# coding: utf-8

from __future__ import annotations
from datetime import date, datetime  # noqa: F401
from enum import Enum, IntEnum

import re  # noqa: F401
from typing import Any, Dict, List, Optional, Type, Union, Callable  # noqa: F401
from pathlib import Path
from typing import TypeVar
Model = TypeVar('Model', bound='BaseModel')
StrBytes = Union[str, bytes]

from pydantic import AnyUrl, BaseModel, EmailStr, Field, validator, root_validator, Extra  # noqa: F401

from dnv_bladed_models.applied_load import AppliedLoad

from dnv_bladed_models.dnv import Dnv

from dnv_bladed_models.environment import Environment

from dnv_bladed_models.event import Event

from dnv_bladed_models.external_module import ExternalModule

from dnv_bladed_models.externally_stepped_simulation import ExternallySteppedSimulation

from dnv_bladed_models.initial_condition import InitialCondition

from dnv_bladed_models.time_domain_outputs import TimeDomainOutputs


class TimeDomainSimulation(Dnv):
    """TimeDomainSimulation - The definition of a time domain analysis - mutually exclusive with &#39;SteadyCalculation&#39;

    Attributes:
        - Duration: float => A number representing a time.  The SI units for time are seconds.
        - LeadInTime: float => A number representing a time.  The SI units for time are seconds.
        - Outputs: TimeDomainOutputs 
        - AdditionalExternalModules: List[ExternalModule] => A list of external module that will be used for this simulation only.  This could be used to apply loading to the structure, or some other load case specific purpose.  If the external module is seeking to simulate an intrinsic property of the turbine, consider moving it into the GlobalExternalModules, or adding it to the Assembly tree.
        - ExternallySteppedSimulation: ExternallySteppedSimulation 
        - Environment: Environment 
        - AppliedLoads: List[AppliedLoad] => A list of point loading definitions which apply a time history of forces to the structure.
        - InitialConditions: List[InitialCondition] => A list of initial conditions to apply at the beginning of the simulation.
        - Events: List[Event] => A list of events that occur during the simulation.
    """

    Duration: Optional[float] = Field(alias="Duration", default=None)
    LeadInTime: Optional[float] = Field(alias="LeadInTime", default=None)
    Outputs: Optional[TimeDomainOutputs] = Field(alias="Outputs", default=None)
    AdditionalExternalModules: Optional[List[ExternalModule]] = Field(alias="AdditionalExternalModules", default=list())
    ExternallySteppedSimulation: Optional[ExternallySteppedSimulation] = Field(alias="ExternallySteppedSimulation", default=None)
    Environment: Optional[Environment] = Field(alias="Environment", default=None)
    AppliedLoads: Optional[List[AppliedLoad]] = Field(alias="AppliedLoads", default=list())
    InitialConditions: Optional[List[InitialCondition]] = Field(alias="InitialConditions", default=list())
    Events: Optional[List[Event]] = Field(alias="Events", default=list())

    class Config:
        extra = Extra.forbid
        validate_assignment = True
        allow_population_by_field_name = True
        pass

    @root_validator(pre=True)
    def parsing_ignores_underscore_properties(cls, values: dict[str, any]):
        allowed_vals = {}
        for key, val in values.items():
            if not key.startswith('_'):
                if isinstance(val, dict):
                    allowed_child_vals = {}
                    for child_key, child_val in val.items():
                        if not child_key.startswith('_'):
                            allowed_child_vals[child_key] = child_val
                    allowed_vals[key] = allowed_child_vals
                else:
                    allowed_vals[key] = val
        return allowed_vals

    def to_json(
        self,
        *, 
        include: Optional[Union['AbstractSetIntStr', 'MappingIntStrAny']] = None, 
        exclude: Optional[Union['AbstractSetIntStr', 'MappingIntStrAny']] = None, 
        by_alias: bool = True, 
        skip_defaults: Optional[bool] = None, 
        exclude_unset: bool = False, 
        exclude_defaults: bool = False, 
        exclude_none: bool = True, 
        encoder: Optional[Callable[[Any], Any]] = None, 
        models_as_dict: bool = True, 
        **dumps_kwargs: Any) -> str:

        r"""
        Generates a JSON string representation of the model.
        
        Notes
        -----
        `include` and `exclude` arguments as per `dict()`.

        `encoder` is an optional function to supply as `default` to json.dumps(), other arguments as per `json.dumps()`.

        Examples
        --------
        >>> model.to_json()

        Renders the full JSON representation of the model object.
        """

        if dumps_kwargs.get('indent') is None:
            dumps_kwargs.update(indent=2)

        return super().json(
                include=include, 
                exclude=exclude, 
                by_alias=by_alias, 
                skip_defaults=skip_defaults, 
                exclude_unset=exclude_unset, 
                exclude_defaults=exclude_defaults, 
                exclude_none=exclude_none, 
                encoder=encoder, 
                models_as_dict=models_as_dict, 
                **dumps_kwargs)
    
    @classmethod
    def from_file(
        cls: Type['Model'],
        path: Union[str, Path]) -> 'Model':
        
        r"""
        Loads a model from a given file path.

        Parameters
        ----------
        path : string
            The file path to the model.

        Returns
        -------
        TimeDomainSimulation
            The model object.

        Raises
        ------
        ValueError, ValidationError
            If the JSON document does not correctly describe the model according to the model schema.

        Examples
        --------
        >>> model = TimeDomainSimulation.from_file('/path/to/file')
        """
        
        return super().parse_file(path=path)
    
    @classmethod
    def from_json(
        cls: Type['Model'],
        b: StrBytes) -> 'Model':

        r"""
        Creates a model object from a JSON string.

        Parameters
        ----------
        b: StrBytes
            The JSON string describing the model.

        Returns
        -------
        TimeDomainSimulation
            The model object.

        Raises
        ------
        ValueError, ValidationError
            If the JSON document does not correctly describe the model according to the model schema.

        Examples
        --------
        >>> model = TimeDomainSimulation.from_json('{ ... }')
        """

        return super().parse_raw(
            b=b,
            content_type='application/json')
    
    @classmethod
    def from_dict(
        cls: Type['Model'],
         obj: Any) -> 'Model':

        r"""
        Creates a model object from a dict.

        Parameters
        ----------
        obj : Any
            The dictionary object describing the model.

        Returns
        -------
        TimeDomainSimulation
            The model object.

        Raises
        ------
        ValueError, ValidationError
            If the JSON document does not correctly describe the model according to the model schema.
        """
        
        return super().parse_obj(obj=obj)
    
    def to_file(
        self,
        path: Union[str, Path]):

        r"""
        Writes the model as a JSON document to a file with UTF8 encoding.        

        Parameters
        ----------                
        path : string
            The file path to which the model will be written.

        Examples
        --------
        >>> model.to_file('/path/to/file')

        """

        with open(file=path, mode='w', encoding="utf8") as output_file:
            output_file.write(self.to_json())

TimeDomainSimulation.update_forward_refs()
